#include <litestep/lsapi/lsapi.h>
#include <assert.h>
#include "vwm.h"
#include "trace.hpp"

static BOOL CALLBACK windowListBuilder(HWND window, LPARAM param);
static bool isMinimized(HWND window);

static HICON getTaskIconWithSize(HWND window, bool big, int timeout);
static HICON getTaskIcon(HWND window, bool preferBig, int timeout);

bool isWindowMinimized(HWND window, const RECT *screenRect)
{
	if(IsIconic(window))
		return true;
	else if(screenRect->left < -10000 && screenRect->top < -10000)
		return true;
	else
		return false;
}

/// Traverse the list of windows and update zOrder and windowsByHandle to
/// reflect any changes which other programs have made. Returns true if a
/// change was found, false otherwise.
bool VWM::updateWindowList()
{
	bool changeFound = false;
	updateNextDraw = false;
	
	zOrder.clear();
	EnumWindows(&windowListBuilder, (LPARAM)&zOrder);
	
	for(map<HWND, WindowData*>::iterator ii=windowsByHandle.begin(); ii!=windowsByHandle.end(); ii++)
	{
		ii->second->focused = false;
		ii->second->visited = false;
	}
	
	vector<WindowData> *ret = new vector<WindowData>();
	int writePos = 0;
	for(unsigned ii=0; ii<zOrder.size(); ii++)
	{
		HWND handle = zOrder[ii];
		if(shouldIgnoreWindow(handle))
			continue;
		zOrder[writePos++] = zOrder[ii];
		
		map<HWND, WindowData*>::iterator jj = windowsByHandle.find(handle);
		WindowData *data;
		if(jj == windowsByHandle.end()) {
			data = noticeWindowCreation(handle);
			windowsByHandle[handle] = data;
			changeFound = true;
		} else {
			data = jj->second;
			if(data->z != ii) {
				data->z = ii;
				changeFound = true;
			}
			if(updateWindow(data)) {
				changeFound = true;
			}
		}
		
		data->visited = true;
	}
	zOrder.resize(writePos);
	
	for(map<HWND, WindowData*>::iterator ii=windowsByHandle.begin(); ii!=windowsByHandle.end(); )
	{
		if(ii->second->visited)
			ii++;
		else
		{
			changeFound = true;
			noticeWindowDisappearance(ii->second);
			delete ii->second;
			ii = windowsByHandle.erase(ii);
		}
	}
	
	HWND newForegroundHandle = GetForegroundWindow();
	do {
		if(windowsByHandle.find(newForegroundHandle) != windowsByHandle.end())
		{
			WindowData *window = windowsByHandle[newForegroundHandle];
			if(!window->isToolWindow && !window->parent) {
				window->focused = true;
				break;
			}
		}
		newForegroundHandle = GetParent(newForegroundHandle);
	} while(newForegroundHandle);
	
	if(newForegroundHandle != foregroundHandle)
	{
		if(windowsByHandle.find(foregroundHandle) != windowsByHandle.end())
			windowsByHandle[foregroundHandle]->focused = false;
		
		changeFound = true;
		foregroundHandle = newForegroundHandle;
	}
	
	return changeFound;
}

bool operator!=(RECT a, RECT b)
{
	return a.left!=b.left || a.right!=b.right || a.top!=b.top || a.bottom!=b.bottom;
}

bool VWM::updateWindow(WindowData *window)
{
	bool changed = false;
	RECT screenPos = {0,0,0,0};
	GetWindowRect(window->handle, &screenPos);
	
	bool minimized = isWindowMinimized(window->handle, &screenPos);
	VirtualDesktop *desk;
	if(minimized) {
		// Don't change a window's assigned desktop while it's minimzed
		desk = window->desk;
	} else {
		desk = deskFromLocation(screenPos);
	}
	
	if(window->minimized != minimized) {
		changed = true;
		window->minimized = minimized;
	}
	if(window->desk != desk) {
		changed = true;
		if(!desk) {
			desk = rescueOffscreenWindow(window->handle, &screenPos);
		}
		if(window->desk)
			window->desk->numTasks--;
		window->desk = desk;
		window->desk->numTasks++;
	}
	if(window->screenPos != screenPos) {
		changed = true;
		window->screenPos = screenPos;
	}
	
	return changed;
}

WindowData *VWM::noticeWindowCreation(HWND handle)
{
	WindowData *ret = new WindowData();
	ret->handle = handle;
	
	RECT screenPos = {0,0,0,0};
	GetWindowRect(handle, &screenPos);
	
	if(isWindowMinimized(handle, &screenPos)) {
		ret->minimized = true;
		ret->desk = currentDesktop;
	} else {
		ret->minimized = false;
		ret->desk = deskFromLocation(screenPos);
		if(!ret->desk)
			ret->desk = rescueOffscreenWindow(handle, &screenPos);
	}
	
	ret->focused = false;
	ret->screenPos = screenPos;
	ret->visited = true;
	
	ret->isToolWindow = (GetWindowLong(handle, GWL_EXSTYLE) & WS_EX_TOOLWINDOW)?true:false;
	ret->parent = GetParent(handle);
	if(!ret->parent)
		ret->parent = GetWindow(handle, GW_OWNER);
	
	ret->smallIcon = getTaskIcon(handle, false, 100);
	ret->bigIcon = getTaskIcon(handle, true, 100);
	
	if(ret->desk)
		ret->desk->numTasks++;
	
	return ret;
}

VirtualDesktop *VWM::rescueOffscreenWindow(HWND handle, RECT *pos)
{
	*pos = getGatherTarget(*pos);
	
	SetWindowPos(handle, NULL,
	             pos->left, pos->top, 0, 0,
	             SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE);
	
	return currentDesktop;
}

void VWM::noticeWindowDisappearance(WindowData *data)
{
	if(data->smallIcon)
		DestroyIcon(data->smallIcon);
	if(data->bigIcon)
		DestroyIcon(data->bigIcon);
	
	if(data->desk)
		data->desk->numTasks--;
}

string getWindowTitle(HWND window)
{
	char buf[1024];
	GetWindowText(window, buf, 1024);
	return string(buf);
}

bool VWM::shouldIgnoreWindow(HWND window)
{
	// Skip hidden windows
	if(!IsWindowVisible(window))
		return true;
	
	// Skip anything that has a parent but isn't a popup
	HWND parent = GetParent(window);
	int windowStyle = GetWindowLong(window, GWL_STYLE);
	if(parent && !(windowStyle & WS_POPUP))
		return true;
	
	// Skip windows which are components of the Litestep desktop (including the VWM itself)
	if(GetWindowLong(window, GWL_USERDATA) == magicDWord)
		return true;

	return false;
}

BOOL CALLBACK windowChildListBuilder(HWND window, LPARAM param)
{
	vector<HWND> *vec = (vector<HWND>*)param;
	vec->push_back(window);
	return TRUE;
}

BOOL CALLBACK windowListBuilder(HWND window, LPARAM param)
{
	vector<HWND> *vec = (vector<HWND>*)param;
	vec->push_back(window);
	
	// If this is a root window, iterate over its children
	HWND parent = GetParent(window);
	if(!parent)
		EnumChildWindows(window, &windowChildListBuilder, param);
	
	return TRUE;
}

void VWM::initTrackingHooks()
{
	// Set a hook to catch WM_WINDOWPOSCHANGED messages.
	// This doesn't seem to work (hook proc doesn't get called)
	// TODO: Fix this
	HWND hookManager = FindWindow("HookMgrClass", NULL);
	if(!hookManager)
		trace << "There is no hook manager!\n";
	SendMessage(hookManager, LM_REGISTERMESSAGE, (WPARAM)WM_WINDOWPOSCHANGED, (LPARAM)hookWindowPosChangedProc);
}
void VWM::cleanupTrackingHooks()
{
}

// The code that sets this hook was commented out, so this can't be terribly important
void hookWindowPosChangedProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	trace << "hookWindowPosChangedProc called.\n";
	LPWINDOWPOS pos = (LPWINDOWPOS)lParam;
	vwm->interceptWindowPosChange(pos->hwnd, pos);
}

void VWM::interceptWindowPosChange(HWND window, WINDOWPOS *pos)
{
	if(shouldIgnoreWindow(window))
		return;
	
	if(windowsByHandle.find(window) == windowsByHandle.end())
	{
		// New window - TODO
	}
	else
	{
		// Changed size or position
		if ( !(pos->flags & SWP_NOSIZE) || !(pos->flags & SWP_NOMOVE) )
		{
			updateWindow(windowsByHandle[window]);
		}
		
		// Changed Z-order
		if ( !(pos->flags & SWP_NOZORDER) )
		{
			// Z-previous window is pos->hwndInsertAfter
			//TODO
		}
	}
}

#define INVALID_ICON(icon) (!(icon) || (int)(icon)==-1 || (int)(icon)==1)

HICON getTaskIcon(HWND window, bool preferBig, int timeout)
{
	HICON icon = NULL;
	
	if(preferBig)
	{
		icon = getTaskIconWithSize(window, true, timeout);
		if(!icon)
			icon = getTaskIconWithSize(window, false, timeout);
	}
	else
	{
		icon = getTaskIconWithSize(window, false, timeout);
		if(!icon)
			icon = getTaskIconWithSize(window, true, timeout);
	}
	
	// If no valid icon was found, use a default icon
	if(INVALID_ICON(icon))
	{
		icon = LoadIcon(NULL, IDI_APPLICATION);
		return icon;
	}
	else
	{
		HICON ret = CopyIcon(icon);
		return ret;
	}
}

HICON getTaskIconWithSize(HWND window, bool big, int timeout)
{
	HICON ret;
	
	int messageIconSize = big ? ICON_BIG : ICON_SMALL;
	if(SendMessageTimeout(
		window,
		WM_GETICON,
		messageIconSize,
		0,
		SMTO_BLOCK|SMTO_ABORTIFHUNG,
		timeout,
		(PDWORD_PTR)&ret))
	{
		if(!INVALID_ICON(ret))
			return ret;
	}
	
	int classLongIconSize = big ? GCL_HICON : GCL_HICONSM;
	ret = (HICON)GetClassLong(window, classLongIconSize);
	if(!INVALID_ICON(ret))
		return ret;

	if(SendMessageTimeout(
		window,
		WM_QUERYDRAGICON,
		messageIconSize,
		0,
		SMTO_BLOCK|SMTO_ABORTIFHUNG,
		timeout,
		(PDWORD_PTR)&ret))
	{
		if(!INVALID_ICON(ret))
			return ret;
	}
	
	return NULL;
}
